# DETAILS:
# - Author: StJohn Piano
# - Date: 2019-02-23
# - Description: This file contains functions useful for dealing with standard Bitcoin transactions, addresses, etc. 


# ENVIRONMENTS
# - Tested on:
# -- Python 2.7.13 running on Mac OS X 10.6.8 (Snow Leopard). 


# NOTES:
#
# - My working definition of a standard transaction: 
# -- It has at least one input and at least one output. 
# -- All input and output addresses are Pay-To-Public-Key-Hash (P2PKH). 
# -- All input scriptSigs contain uncompressed public keys.
#
# - Currently, treat longs as being ints.
# - n is the order of the base point G of the elliptic curve secp256k1. In hex, n = FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFE BAAEDCE6 AF48A03B BFD25E8C D0364141. n is 32 bytes (256 bits) long. The domain of a bitcoin private key is [1,n-1]. 
# - Square brackets [] signify inclusive intervals. 
# - [Hearsay] An ECDSA signature including a low-S value is mathematically equivalent to the same signature that includes the corresponding high-S value. 
# - Currently the Bitcoin network has converged on using low-S values in transaction signatures.
# - low-S domain: [1, floor(n/2)].
# - high-S domain: [floor(n/2)+1, n-1].
# - tx = transaction
# - le = little-endian


# DEPENDENCIES
import bjorn_edstrom_ripemd160 # we use: class RIPEMD160
import ecdsa
import pypy_sha256 # we use: class sha256


# NOTES ON DEPENDENCIES:
# - bjorn_edstrom_ripemd160.py: An implementation of RIPEMD-160 written by <latin-1>B_j_o:r_n_  E_d_s_t_r_o:m_</latin-1>, published 16 December 2007. I downloaded it on 2017-08-04.
# - ecdsa-0.10.tar.gz: ecdsa-0.10.tar.gz is a Python implementation of ECDSA cryptography, written by Peter Pearson, published in 2005. Some comments/updates in the code are dated 2006, 2008, and 2009. Brian Warner added a wrapper around the original code in 2010. I downloaded it in May 2018. 
# - pypy_sha256.py: A Python implementation of SHA256. I don't know who the author is. I downloaded it on 2017-05-12. 








### ALTERATIONS

# make a difficult-to-reverse change to an input


def ensure_low_s_value(s_int):
	# Check that the S-value in the signature is within the low-S domain [1, floor(n/2)]. If it is high-S, alter it to low-S. 
	# high-S domain: [floor(n/2)+1, n-1].
	# [Hearsay] An ECDSA signature including a low-S value is mathematically equivalent to the same signature that includes the corresponding high-S value. 
	n_int = get_secp256k1_n_int()
	limit_int = int(n_int/2)
	validate_int(limit_int)
	validate_int(s_int)
	# is S-value within low-S domain?
	if s_int < 1:
		message = "S value (%d) is less than the minimum permissible value (1)." % s_int
		stop(message)
	elif (1 <= s_int <= limit_int):
		# S-value is in low-S domain
		return s_int
	elif (limit_int < s_int < n_int):
		# S-value is in high-S domain
		# convert the high-S value to a low-S value by using the following equation: S_new = n - S.
		s_new_int = n_int - s_int
		return s_new_int
	else:
		message = "S value (%d) is greater than the maximum permissible value (%d)." % (s_int, n_int)
		stop(message)
	message = "Unknown problem in method ensure_low_s_value()."
	stop(message)


def left_pad_hex_to_length_x(input, x):
	# pad the front with leading zero bytes to reach length = x.
	# 2 hex characters represents 1 byte.
	validate_hex(input)
	n = len(input) / 2 # number of input hex bytes
	if x < n:
		message = "Input is longer than desired padded length. Input: %s" % input
		stop(message)
	m = x - n
	output = input
	if m > 0:
		# add leading zero hex bytes
		output = "00"*m + output
	validate_hex(output)
	return output


def right_pad_hex_to_length_x(input, x):
	# pad the end with leading zero bytes to reach length = x.
	# used for padding little-endian values. 
	# 2 hex characters represents 1 byte.
	validate_hex(input)
	n = len(input) / 2 # number of input hex bytes
	if x < n:
		message = "Input is longer than desired padded length. Input: %s" % input
		stop(message)
	m = x - n
	output = input
	if m > 0:
		# add leading zero hex bytes (at the end)
		output = output + "00"*m
	validate_hex(output)
	return output
	
	
def private_key_hex_to_bitcoin_address(input):
	from binascii import hexlify, unhexlify
	validate_hex_length(input, 32)
	public_key_hex = private_key_hex_to_public_key_hex(input)
	# hash the public key
	hash_hex = public_key_hex_to_hash_hex(public_key_hex)
	# add version byte to public key hash ("00" for "main Bitcoin network").
	hash_hex2 = "00" + hash_hex
	# convert the public key hash to Base58Check format.
	bitcoin_address = hex_to_base58check(hash_hex2)
	return bitcoin_address


def public_key_hex_to_hash_hex(public_key_hex):
	# hash the public key twice: first with SHA256 and then with RIPEMD-160.
	from binascii import hexlify, unhexlify
	validate_hex(public_key_hex)
	digest_hex = sha256(public_key_hex)
	digest2_hex = ripemd160(digest_hex)
	validate_hex_length(digest2_hex, 20)
	return digest2_hex


def private_key_hex_to_public_key_hex(input):
	from binascii import hexlify, unhexlify
	validate_hex_length(input, 32)
	# use ECDSA library to derive the public key from the private key
	input_int = int(input, 16)
	signing_key = ecdsa.SigningKey.from_secret_exponent(input_int, curve=ecdsa.SECP256k1)
	verifying_key = signing_key.verifying_key
	verifying_key_bytes = verifying_key.to_string()
	verifying_key_hex = hexlify(verifying_key_bytes)
	# note: verifying_key.to_string() returns: x_str + y_str
	validate_hex_length(verifying_key_hex, 64)
	# future: is verifying_key_bytes always 64 bytes long? is any padding / checking done explicitly and carefully in the ECDSA library?
	# add identification byte to verifying key ("04" for "uncompressed") to get an uncompressed Bitcoin public key
	public_key_hex = "04" + verifying_key_hex
	validate_hex(public_key_hex)
	return public_key_hex








### GET INFORMATION

# retrieve information or calculate something about the input. 


def hex_len(input):
	# divide hex lengths by 2 to get hex byte length. 
	validate_hex(input)
	n = len(input) / 2
	return n


def count_leading_zero_bytes(input_hex):
	validate_hex(input_hex)
	count = 0
	for i in range(0, len(input_hex), 2):
		byte = input_hex[i:i+2]
		if byte == '00':
			count += 1
		else:
			break
	return count


def get_secp256k1_n_hex():
	# - n is the order of the base point G of the elliptic curve secp256k1. In hex, n = FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFE BAAEDCE6 AF48A03B BFD25E8C D0364141. n is 32 bytes (256 bits) long. The domain of a bitcoin private key is [1,n-1]. 
	n_hex = "FFFFFFFF FFFFFFFF FFFFFFFF FFFFFFFE BAAEDCE6 AF48A03B BFD25E8C D0364141"
	n_hex = n_hex.replace(' ','')
	return n_hex


def get_secp256k1_n_int():
	n_hex = get_secp256k1_n_hex()
	n_int = int(n_hex, 16)
	return n_int


def ripemd160(input_hex):
	from binascii import hexlify, unhexlify
	validate_hex(input_hex)
	input_bytes = unhexlify(input_hex)
	digest_bytes = bjorn_edstrom_ripemd160.RIPEMD160(input_bytes).digest()
	digest_hex = hexlify(digest_bytes)
	validate_hex_length(digest_hex, 20)
	return digest_hex


def sha256(input_hex):
	from binascii import hexlify, unhexlify
	validate_hex(input_hex)
	input_bytes = unhexlify(input_hex)
	digest_bytes = pypy_sha256.sha256(input_bytes).digest()
	digest_hex = hexlify(digest_bytes)
	validate_hex_length(digest_hex, 32)
	return digest_hex

	
def double_sha256(input_hex):
	validate_hex(input_hex)
	digest_hex = sha256(input_hex)
	digest2_hex = sha256(digest_hex)
	validate_hex_length(digest2_hex, 32)
	return digest2_hex


def double_sha256_checksum(input_hex):
	validate_hex(input_hex)
	digest_hex = double_sha256(input_hex)
	checksum_hex = digest_hex[:8] # first 4 bytes
	return checksum_hex








### CONVERSIONS

# convert input from one type to another type. 


def signature_hex_and_public_key_hex_to_scriptSig(signature_hex, public_key_hex):
	# This is my current understanding of a standard scriptSig:
	# 
	# scriptSig:
	# START FORMAT
	# - PUSHDATA: 47 (approximately)
	# - [derived property] PUSHDATA decimal value: 71 (approximately)
	# - signature_data: 71 bytes (approximately)
	# - PUSHDATA: 41
	# - [derived property] PUSHDATA decimal value: 65
	# - public_key_data: 65 bytes ("04", 32-byte big-endian X value, 32-byte big-endian Y value)
	# END FORMAT
	# 
	validate_hex(signature_hex)
	validate_hex_length(public_key_hex, 65)
	signature_hex_pushdata = int_to_var_int(hex_len(signature_hex))
	public_key_hex_pushdata = int_to_var_int(hex_len(public_key_hex))
	scriptSig = signature_hex_pushdata + signature_hex + public_key_hex_pushdata + public_key_hex
	validate_hex(scriptSig)
	script_length = int_to_var_int(hex_len(scriptSig))
	return scriptSig, script_length


def public_key_hash_hex_to_scriptPubKey(hash_hex):
	# This is my current understanding of a standard scriptPubKey:
	# 
	# scriptPubKey:
	# START FORMAT
	# - OP_DUP: 76
	# - OP_HASH160: a9
	# - PUSHDATA: 14
	# - [derived property] PUSHDATA decimal value: 20
	# - public_key_hash: 20 bytes (big-endian)
	# - OP_EQUALVERIFY: 88
	# - OP_CHECKSIG: ac
	# END FORMAT	
	scriptPubKey = "76" + "a9" + "14" + hash_hex + "88" + "ac"
	script_length = hex_len(scriptPubKey)
	script_length = int_to_var_int(script_length)
	return scriptPubKey, script_length


def satoshi_to_tx_output_hex_le(satoshi_amount):
	# convert integer amount to 8 bytes (little-endian). 
	output_hex = int_to_hex(satoshi_amount)
	output_hex_le = reverse_hex_order(output_hex)
	# with little-endian values, have to use right-pad. 
	output_hex_le = right_pad_hex_to_length_x(output_hex_le, 8)
	return output_hex_le
	
	
def tx_index_int_to_tx_index_hex_le(index):
	# previous_output_index is an integer. needs conversion to 4 hex bytes (little endian).
	validate_int(index)
	index_original = index
	index_hex = int_to_hex(index)
	index_hex_le = reverse_hex_order(index_hex)
	# with little-endian values, have to use right-pad. 
	index_hex_le = right_pad_hex_to_length_x(index_hex_le, 4)
	validate_hex_length(index_hex_le, 4)
	return index_hex_le
	

def int_to_var_int(input):
	# Currently only handles 1-byte var_ints.
	# A 1-byte var_int can encode values in the domain [0, 252]. 
	validate_int(input)
	validate_int_value_domain(input, min_value=0, max_value=252)
	if (0 <= input <= 252):
		output = int_to_hex(input)
	else:
		message = "Unknown error in function int_to_var_int()."
		stop(message)
	validate_hex(output)
	return output


def reverse_hex_order(input):
	validate_hex(input)
	x = input
	n = 2 # number of characters per byte
	output = [x[i:i+n] for i in range(0, len(x), n)]
	output.reverse()
	output = "".join(output)
	validate_hex(output)
	return output


def satoshi_to_bitcoin(input):
	# accepts int, returns bitcoin-amount-as-string.
	# Example: accepts 242000, returns "0.00242".
	validate_positive_int(input)
	input_str = str(input)
	n_digits = len(input_str)
	if n_digits <= 8:
		# less than 1 bitcoin.
		# left-pad with zeroes to reach 8 decimal places.
		m = 8 - n_digits
		output = "0"*m + input_str
		# prefix with 0 and decimal point. 
		output = "0." + output
	elif 8 < n_digits:
		output1 = input_str[:-8] # larger part
		output2 = input_str[-8:] # smaller part
		# insert decimal point. 
		output = output1 + "." + output2
	else:
		# future: check for too many satoshi in input. 
		pass
	validate_bitcoin_amount(output)
	return output
	

def bitcoin_to_satoshi(input):
	# accepts bitcoin-amount-as-string, returns int.
	# Example: accepts "0.00242", returns 242000. 
	validate_bitcoin_amount(input)
	# avoid any calculations in floating point. 
	if input.count('.') == 0:
		# no decimal component
		output = input + "0"*8 # add 8 zeros to get satoshi amount.
	elif input.count('.') == 1:
		# increase number of decimal places to 8, then remove the decimal point. 
		input1, input2 = input.split('.')
		n = len(input2)
		m = 8 - n # number of "missing" decimal places. 
		output = input.replace('.','') + "0"*m
	else:
		message = "bitcoin amount should contain 0 or 1 decimal points."
		stop(message)
	# remove any leading zeros
	output = output.lstrip('0')
	# convert to integer
	output = int(output)
	validate_positive_int(output)
	return output


def private_key_hex_to_wif(input):
	# convert the private key to WIF format.
	# add a "80" byte at the front (to indicate Bitcoin "mainnet").
	validate_hex(input)
	input = "80" + input
	private_key_wif = hex_to_base58check(input)
	return private_key_wif


def bitcoin_address_to_public_key_hash_hex(input):
	# My working definition of a standard address is: Pay-To-Public-Key-Hash (P2PKH) where the public key was not compressed before being hashed. 
	output = base58check_to_hex(input)
	# The first byte in the result is the version byte. For the main Bitcoin network, the version byte is 0x00. Remove it. 
	# future: confirm that it is 0x00. 
	output = output[2:]
	# The final result should be the public key hash in hex bytes. 
	# public key hash = RIPEMD-160(SHA256(public_key)) and should be exactly 20 bytes long. 
	validate_hex_length(output, 20)
	return output
	

def base58check_to_hex(input):
	validate_string(input)
	# 1) Count the number of leading '1' characters in the address, then remove them. 
	input2 = input.lstrip('1')
	count_1s = len(input) - len(input2)
	# 2) Convert the result of step (1) from the Bitcoin base-58 encoding to a base-10 integer.
	output = base58_to_int(input2)
	# 3) Convert the result of step (2) from a base-10 integer to a hex byte string.
	output = int_to_hex(output)
	# 4) The last four bytes in the result of step (3) are the checksum. Remove the checksum and save its value.
	checksum = output[-8:]
	output2 = output[:-8]
	# 5) Let the result of step (1) be N. Add N leading zero bytes to the result of step (4).
	output3 = (count_1s * '00') + output2
	# calculate the checksum and confirm that the saved value is correct. 
	checksum_new = double_sha256_checksum(output3)
	if checksum != checksum_new:
		message = "Checksum value does not match calculated checksum."
		message += "\nInput: %s" % input
		message += "\nInput hex: %s" % output
		message += "\nInput hex without checksum: %s" % output2
		message += "\nInput hex without checksum and with zeroes: %s" % output3
		message += "\nChecksum value     : %s" % checksum
		message += "\nCalculated checksum: %s" % checksum_new
		stop(message)
	return output3
	
	
def hex_to_base58check(input_hex):
	validate_hex(input_hex)
	checksum_hex = double_sha256_checksum(input_hex)
	item_hex = input_hex + checksum_hex
	item_base58_str = hex_to_base58(item_hex)
	n = count_leading_zero_bytes(input_hex)
	output = n * '1' + item_base58_str
	return output


def hex_to_base58(input_hex):
	validate_hex(input_hex)
	input_int = int(input_hex, 16)
	# the string base58_symbols can accessed as a 0-indexed list of characters.
	base58_symbols = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
	output_str = ''
	item_int = input_int
	while item_int > 0:
		item_int, remainder = divmod(item_int, 58)
		# use remainder as index for accessing the corresponding base58 symbol. 
		output_str += base58_symbols[remainder]
	output_str = ''.join(reversed(output_str))
	return output_str
	

def base58_to_int(input):
	# convert a string from Bitcoin base-58 encoding to a base-10 integer. 
	# the string base58_symbols can accessed as a 0-indexed list of characters.
	base58_symbols = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
	n = len(input)
	m = n - 1
	output = 0 # running total
	for i in range(n):
		character = input[i]
		value = base58_symbols.index(character)
		result = value * (58 ** (m - i))
		output += result
	validate_int(output)
	return output


def bytes_to_hex(input):
	from binascii import hexlify
	validate_bytes(input)
	output = hexlify(input)
	validate_hex(output)
	return output


def int_to_hex(input):
	# accepts integer, returns hex byte sequence (as a string)
	validate_int(input)
	output = hex(input)[2:] # remove 0x prefix.
	# Python-specific: remove L suffix if it exists
	if output[-1] == "L":
		output = output[:-1]
	# ensure an even number of characters. 1 byte == 2 hex characters. An extra hex character by itself is a half-byte. 
	if len(output) % 2 == 1:
		output = "0" + output # add a leading 0 (a single empty half-byte in hex)
	validate_hex(output)
	return output
	
	
	

	



### VALIDATIONS

# raise error if input is an invalid value. 


def validate_list_contains_items(input, items):
	# input is a list. items is a list.
	validate_list(input)
	validate_list(items)
	not_found = []
	for x in items:
		if x not in input:
			not_found.append(x)
	if len(not_found) == 0:
		return True
	message = "Some items are not in the list."
	message += "\nList items:"
	for x in input:
		message += "\n- %s" % str(x)
	message += "\nItems not found:"
	for x in not_found:
		message += "\n- %s" % str(x)
	stop(message)


def validate_item_in_list(input, list1):
	validate_list(list1)
	if input in list1:
		return True
	message = "Input not in list. Input: %s" % str(input)
	message += "\nList items:"
	for item in list1:
		message += "\n- %s" % str(item)
	stop(message)


def validate_public_key_hash_hex(input):
	n = len(input) / 2 # number of bytes
	if n == 20:
		return True
	else:
		message = "Byte length of public key hash is not 20."
		message += "\nPublic key hash: %s" % input
		message += "\nByte length of public key hash: %d" % n
		stop(message)
	message = "Unknown problem in function validate_public_key_hash_hex()."
	stop(message)


def validate_bitcoin_private_key_bytes(input):
	from binascii import hexlify
	# Currently, only printable ASCII bytes are permitted. 
	validate_bytes(input)
	if not is_printable_ascii(input):
		message = "Input does not consist only of printable ASCII characters. Input: %s" % repr(input)
		stop(message)
	# A private key must be at least 1 byte long and at most 32 bytes long. 
	validate_bytes_length_domain(input, min_length=1, max_length=32)
	# Convert to hex and use existing validation function
	input_hex = hexlify(input)
	validate_bitcoin_private_key_hex(input_hex)


def validate_bitcoin_private_key_hex(input):
	validate_hex(input)
	# A private key must be at least 1 byte long and at most 32 bytes long. 
	validate_hex_length_domain(input, min_length=1, max_length=32)
	input_int = int(input, 16)
	# secp256k1 elliptic curve properties:
	n_int = get_secp256k1_n_int()
	max_int = n_int - 1
	min_int = 1
	validate_int_value_domain(input_int, min_int, max_int)


def validate_bytes_length(input, x):
	# byte length of hex input must be x. 
	validate_bytes(input)
	validate_int(x)
	n = len(input) # number of bytes
	validate_int_value(n, x)


def validate_bytes_length_domain(input, min_length, max_length):
	validate_bytes(input)
	n = len(input) # number of bytes
	validate_int_value_domain(n, min_length, max_length)


def validate_hex_length(input, x):
	# byte length of hex input must be x. 
	validate_hex(input)
	validate_int(x)
	# 2 hex characters represent 1 byte.
	n = len(input) / 2 # number of hex bytes
	validate_int_value(n, x)
	
	
def validate_int_value(input, x):
	# input value must be x. 
	validate_int(input)
	validate_int(x)
	if input == x:
		return True
	else:
		message = "Input is not exactly equal to %d." % x
		message += "\nInput: %d" % input
		stop(message)
	message = "Unknown problem in function validate_int_value()."
	stop(message)


def validate_hex_length_domain(input, min_length, max_length):
	validate_hex(input)
	# 2 hex characters represent 1 byte.
	n = len(input) / 2 # number of hex bytes
	validate_int_value_domain(n, min_length, max_length)
	

def validate_int_value_domain(n, min_value, max_value):
	# n must be within [min_value, max_value].
	validate_int(n)
	validate_int(min_value)
	validate_int(max_value)
	if min_value > max_value:
		message = "Minimum value must be less than maximum value."
		message += "\nMininimum value: %d" % min_value
		message += "\nMaximum value: %d" % max_value
		stop(message)
	if (min_value <= n <= max_value):
		return True
	elif n < min_value:
		message = "Input is less than the minimum value."
		message += "\nMinimum value: %d" % min_value
		message += "\nInput: %s" % n
		stop(message)
	elif max_value < n:
		message = "Input is greater than the maximum value."
		message += "\nMaximum value: %d" % max_value
		message += "\nInput: %s" % n
		stop(message)
	message = "Unknown problem in function validate_int_value_domain()."
	stop(message)


def validate_bitcoin_amount(input):
	# checks format. accepts string e.g. "0.00242"
	# future: investigate accuracy of float(). 
	validate_string(input)
	max_amount = 21*(10**6) # 21 million bitcoin
	min_amount = 0.00000001 # 1 satoshi
	# input can contain 0 or 1 period. 
	if input.count('.') not in [0, 1]:
		message = "Input can contain either 1 or 0 periods ('.'). Input: %s" % input
		stop(message)
	# apart from the period, input can contain only digits. 
	input2 = input.replace('.','')
	if not is_digits(input2):
		message = "Apart from an optional period ('.'), input can contain only digits. Input: %s" % input
		stop(message)
	# the numerical value must be within the acceptable domain. 
	if float(input) < float(min_amount):
		message = "The minimum input amount is %f. Input: %s" % (min_amount, input)
		stop(message)
	elif float(input) > float(max_amount):
		message = "The maximum input amount is %f. Input: %s" % (max_amount, input)
		stop(message)
	else:
		# input is within acceptable domain. 
		return True
	message = "Unknown problem in function validate_bitcoin_amount()."
	stop(message)


def validate_list(input):
	if isinstance(input, list):
		return True
	message = "Input is not a list. Input type is: %s" % type(input)
	stop(message)


def validate_hex(input):
	validate_string(input)
	if not is_hex(input):
		message = "Input must consist only of hex characters. Input: %s" % input
		stop(message)
	# 2 hex characters represents 1 byte, so input must have an even number of characters.
	if len(input) % 2 == 0:
		return True
	else:
		message = "Input must have an even number of hex characters. Input: %s" % input
		stop(message)
	message = "Unknown problem in function validate_hex()."
	stop(message)
	

def validate_string_is_digits(input):
	validate_string(input)
	if not is_digits(input):
		message = "Input contains non-digit characters. Input: %s" % input
		stop(message)
	
	
def validate_positive_int(input):
	validate_int(input)
	if input >= 0:
		return True
	message = "Input is negative. Input: %d" % input
	stop(message)


def validate_int(input):
	if isinstance(input, int):
		return True
	# treating longs as ints: can this cause problems?
	if isinstance(input, long):
		return True
	message = "Input is not an integer. Input type is: %s" % type(input)
	stop(message)


def validate_bytes(input):
	validate_string(input)


def validate_string(input):
	if isinstance(input, str):
		return True
	message = "Input is not a string. Input type is: %s" % type(input)
	stop(message)








### CHECKS

# return True or False. 


def is_hex(input):
	hex_characters = "0123456789abcdef"
	for c in input: # c = character
		if c not in hex_characters:
			return False
	return True


def is_digits(input):
	digits = "0123456789"
	for c in input: # c = character
		if c not in digits:
			return False
	return True


def is_printable_ascii(input):
	characters = "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~"
	whitespace = "\n\t "
	byte_values = characters + whitespace
	for c in input: # c = character
		if c not in byte_values:
			return False
	return True








### ERRORS


def stop(message):
	# raise an Exception to get a traceback. 
	raise Exception("\n\nERROR: %s\n" % message)


