#!/opt/local/bin/python


# DETAILS:
# - Author: StJohn Piano
# - Date: 2019-02-23
# - Description: This script creates and signs a nonstandard Bitcoin transaction with one or more standard p2pkh inputs and one or more outputs. Outputs can be standard p2pkh outputs or nonstandard p2sh outputs.


# ENVIRONMENTS
# - Tested on:
# -- Python 2.7.13 running on Mac OS X 10.6.8 (Snow Leopard). 


# NOTES:
#
# - All input variables for this script are strings. If any inputs are not to be used, they can be set to the empty string {""}. 
#
# - The change address is more accurately a single output sent to that address. 
# 
# - In input_data, value can be included as either satoshi (e.g. "satoshi_amount": "241777") or bitcoin (e.g. "bitcoin_amount": "0.00241777"). 
# 
# - p2pkh = Pay-To-Public-Key-Hash (standard input / output type)
# - p2sh = Pay-To-Script-Hash (nonstandard, can be used for multisignature)


# DEPENDENCIES
import bitcoin_functions_2 as bf
import transaction_3 as tr
import nonstandard_bitcoin_functions_2 as nbf
import nonstandard_transaction_3 as ntr
import math
from collections import defaultdict




def main():
	
	
	##### START CONTROLS
	random_values = [
		"",
	]
	# random_value must be between 1 and 32 bytes. If random_value_type is "raw_bytes", then random_value must be between 1 and 32 ASCII characters. If random_value_type is "hex_bytes", then random_value must be an even number of hex characters and between 2 and 64 hex characters. 
	# Note: Every ECDSA signature (one for each input in a transaction) requires new random entropy. 
	# random_value_type options: ["raw_bytes", "hex_bytes"]
	random_value_type = "hex_bytes"
	input_data = [
		{
		"txid": "",
		"previous_output_index": "",
		"private_key_hex": "",
		"satoshi_amount": "", 
		#"bitcoin_amount": "", 
		"input_type": "p2pkh", 
		},
		"txid": "",
		"previous_output_index": "",
		"private_key_hex": "",
		"satoshi_amount": "", 
		#"bitcoin_amount": "", 
		"input_type": "p2pkh", 
		},
	]
	output_data = [
		{
		"address": "",
		"satoshi_amount": "",
		#"bitcoin_amount": "",
		"output_type": "p2sh", 
		},
		{
		"address": "",
		"satoshi_amount": "",
		#"bitcoin_amount": "",
		"output_type": "p2sh", 
		},
	]
	change_address = ""
	# note: the fee will be subtracted from the amount that is being sent to the change address.
	fee = "" # satoshi
	fee_rate = "" # satoshi / byte
	# fee_type options: ["fee", "fee_rate"]
	fee_type = "fee_rate"
	##### END CONTROLS
	
	
	
	
	##### START SETTINGS
	
	random_value_type_options = ["raw_bytes", "hex_bytes"]
	fee_type_options = ["fee", "fee_rate"]
	input_type_options = ["p2pkh"]
	output_type_options = ["p2pkh", "p2sh"]
	
	##### END SETTINGS
	
	
	
	
	# preparation
	# some values will be converted to integers. 
	# - fee, input["previous_output_index"], input["satoshi_amount"], output["satoshi_amount"]
	# some values will be converted to floats:
	# - fee_rate
	
	bf.validate_string(random_value_type)
	bf.validate_string(change_address)
	bf.validate_string(fee)
	bf.validate_string(fee_rate)
	bf.validate_string(fee_type)
	
	bf.validate_item_in_list(fee_type, fee_type_options)
	if fee_type == "fee":
		bf.validate_string_is_digits(fee)
		fee = int(fee)
	# future: validate that fee_rate is float or integer. 
	elif fee_type == "fee_rate":
		fee_rate = float(fee_rate)
		if fee_rate < 0:
			message = "Fee rate is negative. Fee rate: %s" % str(fee_rate)
			stop(message)
	
	bf.validate_item_in_list(random_value_type, random_value_type_options)
	if random_value_type == "raw_bytes":
		for random_value in random_values:
			bf.validate_bytes_length_domain(random_value, 1, 32)
	elif random_value_type == "hex_bytes":
		for random_value in random_values:
			bf.validate_hex_length_domain(random_value, 1, 32)
	
	
	# id = "input data"
	for id_item in input_data:
		index = id_item["previous_output_index"]
		bf.validate_string_is_digits(index)
		id_item["previous_output_index"] = int(index)
		bf.validate_item_in_list(id_item["input_type"], input_type_options)
		if "bitcoin_amount" in id_item:
			b = id_item["bitcoin_amount"]
			bf.validate_bitcoin_amount(b)
			id_item["satoshi_amount"] = bf.bitcoin_to_satoshi(b)
		elif "satoshi_amount" in id_item:
			s = id_item["satoshi_amount"]
			bf.validate_string_is_digits(s)
			s = int(s)
			id_item["satoshi_amount"] = s
			id_item["bitcoin_amount"] = bf.satoshi_to_bitcoin(s)
	
	
	# od = "output data"
	for od_item in output_data:
		bf.validate_item_in_list(od_item["output_type"], output_type_options)
		if "bitcoin_amount" in od_item:
			b = od_item["bitcoin_amount"]
			bf.validate_bitcoin_amount(b)
			od_item["satoshi_amount"] = bf.bitcoin_to_satoshi(b)
		elif "satoshi_amount" in od_item:
			s = od_item["satoshi_amount"]
			bf.validate_string_is_digits(s)
			s = int(s)
			od_item["satoshi_amount"] = s
			od_item["bitcoin_amount"] = bf.satoshi_to_bitcoin(s)
	
	
	
	# after all changes: final round of validation. 
	
	for id_item in input_data:
		bf.validate_list_contains_items(id_item.keys(), items=["txid", "previous_output_index", "private_key_hex", "satoshi_amount", "bitcoin_amount", "input_type"])
		bf.validate_hex_length(id_item["txid"], 32)
		bf.validate_positive_int(id_item["previous_output_index"])
		bf.validate_bitcoin_private_key_hex(id_item["private_key_hex"])
		bf.validate_positive_int(id_item["satoshi_amount"])
		bf.validate_bitcoin_amount(id_item["bitcoin_amount"])
	
	# future: validate_bitcoin_address() for the output addresses and the change address. 
	
	for od_item in output_data:
		bf.validate_list_contains_items(od_item.keys(), items=["address", "satoshi_amount", "bitcoin_amount", "output_type"])
		bf.validate_string(od_item["address"])
		bf.validate_positive_int(od_item["satoshi_amount"])
		bf.validate_bitcoin_amount(od_item["bitcoin_amount"])
	
	
	
	
	print "\n### START CREATION OF NONSTANDARD BITCOIN TRANSACTION\n"
	
	
	print "- Fee type: %s" % fee_type
	
	if fee_type == "fee":
		print "- Fee: %s (satoshi)" % fee
	elif fee_type == "fee_rate":
		print "- Fee rate: %s (satoshi / byte)" % fee_rate
	
	
	# if random_value type is "raw_bytes", convert to hex.
	if random_value_type == "raw_bytes":
		random_values = [bf.bytes_to_hex(value) for value in random_values]
	
	
	# create inputs
	inputs = []
	for id_item in input_data:
		if id_item["input_type"] == "p2pkh":
			input = tr.Input.create_from_txid(
				txid = id_item["txid"],
				previous_output_index = id_item["previous_output_index"],
				private_key_hex = id_item["private_key_hex"],
				bitcoin_amount = id_item["bitcoin_amount"],
			)
		inputs.append(input)
	
	
	# create outputs
	outputs = []
	for od_item in output_data:
		if od_item["output_type"] == "p2pkh":
			output = tr.Output.create_from_address(
				address = od_item["address"],
				bitcoin_amount = od_item["bitcoin_amount"],
			)
		elif od_item["output_type"] == "p2sh":
			output = ntr.P2SHOutput.create_from_address(
				address = od_item["address"],
				bitcoin_amount = od_item["bitcoin_amount"],
			)
		outputs.append(output)
	
	print "- Number of inputs (i.e. as-yet-unspent outputs): %d" % len(inputs)
	print "- Number of outputs: %d" % len(outputs)
	
	
	print "- Change address: %s" % change_address
	# find the change_output corresponding to the change_address.
	# future: how to handle the case where multiple transaction outputs have the change_address?
	change_output = None
	for output in outputs:
		if output.address == change_address:
			change_output = output
			break
	if change_output == None:
		message = "Change address not found in output addresses."
		message += "\nOutput addresses:"
		addresses = [o.address for o in outputs]
		for address in addresses:
			message += "\n- %s" % address
		message += "\nChange address: %s" % change_address
		stop(message)
		
	# Display the value to be sent to the change address. 
	print "- Amount to be sent to the change address: %s" % change_output.bitcoin_amount
	
	
	# print out input addresses and the total value being spent from each one. 
	# future: create and test a transaction that spends two inputs from the same address.
	input_addresses = defaultdict(int)
	for input in inputs:
		value = input.satoshi_amount
		address = input.address
		input_addresses[address] += value
	print "- Input addresses, with total-value-to-be-sent:"
	for address in input_addresses:
		satoshi = input_addresses[address]
		bitcoin = bf.satoshi_to_bitcoin(satoshi)
		print "-- %s: %s" % (address, bitcoin)
	
	# print out output addresses and the total value being sent to each one. 
	# future: create and test a transaction that sends two outputs to the same address.
	output_addresses = defaultdict(int)
	for output in outputs:
		value = output.satoshi_amount
		address = output.address
		output_addresses[address] += value
	print "- Output addresses, with total-value-to-be-received:"
	for address in output_addresses:
		satoshi = output_addresses[address]
		bitcoin = bf.satoshi_to_bitcoin(satoshi)
		print "-- %s: %s" % (address, bitcoin)
	
	
	# confirm that total input value >= total output value
	input_total = sum([i.satoshi_amount for i in inputs])
	output_total = sum([o.satoshi_amount for o in outputs])
	print "- Total value of all inputs: %s" % bf.satoshi_to_bitcoin(input_total)
	print "- Total value of all outputs: %s" % bf.satoshi_to_bitcoin(output_total)
	
	
	# send any extra input value (i.e. not assigned to be sent to an output) to the change_output. 
	# these amounts are in satoshis. 
	if input_total == output_total:
		print "- Total value of all inputs exactly matches total value of all outputs."
	elif input_total > output_total:
		extra = input_total - output_total
		new_amount = change_output.satoshi_amount + extra
		change_output.set_satoshi_amount(new_amount)
		print "- Total value of all inputs is greater than total value of all outputs."
		print "-- Some input value has not been assigned to an output."
		print "-- Extra value: %s" % bf.satoshi_to_bitcoin(extra)
		print "-- This extra value will be sent to the change address."
		print "-- Change address: %s" % change_output.address
		print "-- New amount to be sent to change address: %s" % change_output.bitcoin_amount
	elif input_total < output_total:
		difference = output_total - input_total
		message = "Total value of all inputs is less than total value of all outputs."
		message += "\n- Difference: %s" % bf.satoshi_to_bitcoin(difference)
		stop(message)
	
	
	
	
	# estimate size of transaction
	
	# signed transaction contains:
	# - version (4 bytes)
	# - input_count (1-byte var_int) [1 byte if number of inputs is small]
	# - concatenated inputs (with signatures)
	# - output_count (1-byte var_int) [1 byte if number of outputs is small]
	# - concatenated outputs
	# - block lock time (4 bytes)
	
	# P2PKH input contains: 
	# - previous_output_hash (32 bytes)
	# - previous_output_index (4 bytes)
	# - script_length (1-byte var_int)
	# - scriptSig [approximately 138 bytes]
	# - sequence (4 bytes)
	
	# P2PKH output contains:
	# - value (8 bytes)
	# - script_length (1-byte var_int)
	# - scriptPubKey (25 bytes)
	
	# P2SH output contains:
	# - value (8 bytes)
	# - script_length (1-byte var_int)
	# - scriptPubKey (23 bytes)
	
	# 23: byte length of a P2SH scriptPubKey (containing a redeemScript hash).
	# 25: byte length of a P2PKH scriptPubKey (containing a public key hash)
	
	# approximate size of a transaction with 1 P2PKH input and 1 P2SH output: 221 bytes
	
	input_count_size = bf.hex_len(bf.int_to_var_int(len(inputs)))
	inputs_size_total = len(inputs) * (32 + 4 + 1 + 138 + 4) # bytes
	
	output_count_size = bf.hex_len(bf.int_to_var_int(len(outputs)))
	
	# distinguish between p2pkh and p2sh outputs. 
	output_size_total = 0 # bytes
	for output in outputs:
		if hasattr(output, "output_type"):
			if output.output_type == "p2sh":
				output_size = 8 + 1 + 23
		else:
			# p2pkh output
			output_size = 8 + 1 + 25
		output_size_total += output_size
		
	estimated_tx_size = (
		4 + 
		input_count_size + 
		inputs_size_total + 
		output_count_size + 
		output_size_total + 
		4
		)
	
	print "- Estimated transaction size: %d bytes" % estimated_tx_size
	
	
	
	
	# calculate approximate fee
	final_fee = None
	if fee_type == "fee":
		final_fee = fee
		print "- Final fee: %d (satoshi)" % final_fee
	elif fee_type == "fee_rate":
		final_fee = estimated_tx_size * fee_rate
		# round up to the nearest satoshi. 
		final_fee = int(math.ceil(float(final_fee)))
		print "- Fee rate: %s (satoshi / byte)" % str(fee_rate)
		print "- Calculate %d * %s and round up to nearest satoshi." % (estimated_tx_size, str(fee_rate))
		print "- Final fee: %d (satoshi)" % final_fee
		
	if final_fee == None:
		message = "Final fee not calculated."
		stop(message)
		
	print "- Final fee rate (using estimated transaction size): %.4f (satoshi per byte)" % (float(final_fee) / estimated_tx_size)
	
	# check whether the fee can be paid from the amount assigned to the change address. 
	if final_fee > change_output.satoshi_amount:
		message = "The fee is greater than the satoshi amount that is assigned to the change address."
		message += "\nFee: %d (satoshi)" % final_fee
		message += "\nAmount to be sent to change address: %d (satoshi)" % change_output.satoshi_amount
		message += "\nDifference: %d (satoshi)" % (final_fee - change_output.satoshi_amount)
		stop(message)
	
	# subtract fee from value assigned to change_output
	original_amount = change_output.satoshi_amount
	new_amount = change_output.satoshi_amount - final_fee
	change_output.set_satoshi_amount(new_amount)
	print "- Fee subtracted from amount to be sent to change address."
	print "- New amount to be sent to change address: %d (satoshi)" % change_output.satoshi_amount
	
	
	
	
	# create transaction
	tx = ntr.NonstandardTransaction.create(inputs, outputs)
	
	
	# sign transaction
	tx.sign(random_values)
	
	
	
	
	# calculate actual size and actual fee rate
	signed_tx_size = bf.hex_len(tx.get_signed_form())
	print "- Size of signed transaction: %d bytes" % signed_tx_size
	print "- (signed_tx_size - estimated_tx_size): %d bytes" % (signed_tx_size - estimated_tx_size)
	signed_tx_fee_rate = float(final_fee) / signed_tx_size
	print "- Fee rate in signed transaction: %s (satoshi / byte)" % str(round(signed_tx_fee_rate,3))
	
	
	# display inputs, outputs, and transaction forms
	
	for i, input in enumerate(tx.inputs):
		print ""
		print "Input %d:" % i
		print input.get_description()
		
	for i, output in enumerate(tx.outputs):
		print ""
		if hasattr(output, "output_type"):
			if output.output_type == "p2sh":
				print "Output %d (P2SH):" % i
		else:
			print "Output %d:" % i
		print output.get_description()
	
	print ""
	print tx.get_description_of_unsigned_form()
	
	for i in xrange(len(inputs)):
		print ""
		print tx.get_description_of_signable_form(input_index=i)
	
	print ""
	print tx.get_description_of_signed_form()
	
	print ""
	print "Signed transaction:"
	print tx.get_signed_form()
	
	
	print "\n### END CREATION OF BITCOIN TRANSACTION\n"




def stop(message):
	bf.stop(message)


if __name__ == '__main__': main()