# DETAILS:
# - Author: StJohn Piano
# - Date: 2018-10-05
# - Description: This file contains functions useful for dealing with nonstandard Bitcoin transactions, addresses, etc. It relies on functions in bitcoin_functions_2.py. 


# ENVIRONMENTS
# - Tested on:
# -- Python 2.7.13 running on Mac OS X 10.6.8 (Snow Leopard). 


# NOTES:
#
# - The address format of P2SH addresses is:
# START FORMAT
# base58-encode( [one-byte version] + [20-byte hash of redeemScript] + [4-byte checksum] )
# END FORMAT
# On the main network, the version byte is 0x05.
#
# - The P2SH outputScript is: 
# START FORMAT
# OP_HASH160 PUSHDATA(20) [20-byte op_hash160 hash of redeemScript] OP_EQUAL
# END FORMAT
#
# - OP_HASH160: a9
# - PUSHDATA(20): 14
# - OP_EQUAL: 87



import bitcoin_functions_2 as bf









### CONVERSIONS

# convert input from one type to another type. 



def redeem_script_hash_hex_to_script(input):
	# - The P2SH outputScript is: 
	# START FORMAT
	# - OP_HASH160: a9
	# - PUSHDATA(20): 14
	# - [derived property] PUSHDATA decimal value: 20
	# - redeem_script_hash: 20 bytes (big-endian) 
	# - OP_EQUAL: 87
	# END FORMAT
	bf.validate_hex_length(input, 20)
	script = "a9" + "14" + input + "87"
	script_length = bf.hex_len(script)
	script_length = bf.int_to_var_int(script_length)
	return script, script_length
	
	


def p2sh_address_to_redeem_script_hash_hex(input):
	validate_p2sh_address(input)
	output = bf.base58check_to_hex(input)
	# output: 1-byte version + 20-byte script hash.
	script_hash = output[2:] # last 20 bytes
	bf.validate_hex_length(script_hash, 20)
	return script_hash
	
	




### VALIDATION

# raise error if input is an invalid value. 


def validate_p2sh_address(input):
	bf.validate_string(input)
	# future: are p2sh addresses all the same length? if so, run a validate_string_length() on the input. 
	output = bf.base58check_to_hex(input)
	# output should be 1-byte version + 20-byte script hash. 
	bf.validate_hex_length(output, 21)
	version = output[:2] # first byte
	if version != "05":
		message = "Byte 0 of input is not 0x05. Byte 0 value: %s" % version
		message += "\nInput: %s" % input
		message += "\nInput (decoded from base58check): %s" % output
		bf.stop(message)
	# there's nothing to validate about a hash apart from its length (which was implicitly verified earlier). 
	# script_hash = output[2:] # last 20 bytes
	

















