# DETAILS
# - Author: StJohn Piano
# - Date: 2019-02-23
# - Description: This file contains classes for a nonstandard Bitcoin transaction (i.e. contains at least one P2SH output) and a nonstandard Bitcoin transaction output (P2SH). 


# ENVIRONMENTS
# - Tested on:
# -- Python 2.7.13 running on Mac OS X 10.6.8 (Snow Leopard).


# NOTES:
# - The nonstandard Bitcoin transaction can still have one or more standard (P2PKH) outputs. 
# - I think the P2SH script might still be called scriptPubKey in the original bitcoind code. 
# - P2SHOutput sends bitcoin to hashes of redeemScripts. 


# DEPENDENCIES
import bitcoin_functions_2 as bf
import ecdsa
import nonstandard_bitcoin_functions_2 as nbf






class NonstandardTransaction():


	def __init__(self):
		# properties that go into the raw transaction:
		self.version = "01000000"
		self.input_count = None # var_int
		self.inputs = []
		self.output_count = None # var_int
		self.outputs = []
		self.block_lock_time = "00000000"
		# properties that go into the raw transaction-in-signable-form:
		self.hash_type_4_byte = "01000000" # 4 bytes (little-endian)
		# properties that go into the signed transaction:
		self.hash_type_1_byte = "01"
	
	
	def get_unsigned_form(self):
		# This is the basic transaction, with no signatures.
		s = self.version
		s += self.input_count
		for i, input in enumerate(self.inputs):
			s += input.previous_output_hash
			s += input.previous_output_index
			# input.script_length not included
			# input.scriptSig not included
			s += input.sequence
		s += self.output_count
		for output in self.outputs:
			# distinguish between p2pkh and p2sh outputs. 
			if hasattr(output, "output_type"):
				if output.output_type == "p2sh":
					s += output.value
					s += output.script_length
					s += output.script
			else:
				# p2pkh output.
				s += output.value
				s += output.script_length
				s += output.scriptPubKey
		s += self.block_lock_time
		return s
	
	
	def get_description_of_unsigned_form(self):
		s = "Nonstandard Transaction (unsigned form): {%d bytes}" % bf.hex_len(self.get_unsigned_form())
		s += "\n- version: %s (%d bytes)" % (self.version, bf.hex_len(self.version))
		s += "\n- input_count: %s (%d bytes)" % (self.input_count, bf.hex_len(self.input_count))
		for i, input in enumerate(self.inputs):
			s += "\n- Input %d: {%d bytes}" % (i, bf.hex_len(input.get_unsigned_form()))
			s += "\n-- previous_output_hash: %s (%d bytes)" % (input.previous_output_hash, bf.hex_len(input.previous_output_hash))
			s += "\n-- previous_output_index: %s (%d bytes)" % (input.previous_output_index, bf.hex_len(input.previous_output_index))
			s += "\n-- script_length: %s (%d bytes)" % ("[none]", 0)
			s += "\n-- scriptSig: %s (%d bytes)" % ("[none]", 0)
			s += "\n-- sequence: %s (%d bytes)" % (input.sequence, bf.hex_len(input.sequence))
		s += "\n- output_count: %s (%d bytes)" % (self.output_count, bf.hex_len(self.output_count))
		for i, output in enumerate(self.outputs):
			# distinguish between p2pkh and p2sh outputs. 
			if hasattr(output, "output_type"):
				if output.output_type == "p2sh":
					s += "\n- Output %d (P2SH): {%d bytes}" % (i, bf.hex_len(output.get_form()))
					s += "\n-- value: %s (%d bytes)" % (output.value, bf.hex_len(output.value))
					s += "\n-- script_length: %s (%d bytes)" % (output.script_length, bf.hex_len(output.script_length))
					s += "\n-- script: %s (%d bytes)" % (output.script, bf.hex_len(output.script))
			else:
				# p2pkh output.
				s += "\n- Output %d: {%d bytes}" % (i, bf.hex_len(output.get_form()))
				s += "\n-- value: %s (%d bytes)" % (output.value, bf.hex_len(output.value))
				s += "\n-- script_length: %s (%d bytes)" % (output.script_length, bf.hex_len(output.script_length))
				s += "\n-- scriptPubKey: %s (%d bytes)" % (output.scriptPubKey, bf.hex_len(output.scriptPubKey))
		s += "\n- block_lock_time: %s (%d bytes)" % (self.block_lock_time, bf.hex_len(self.block_lock_time))
		return s


	@classmethod
	def create(cls, inputs, outputs):
		t = NonstandardTransaction()
		t.input_count = bf.int_to_var_int(len(inputs))
		t.inputs = inputs
		t.output_count = bf.int_to_var_int(len(outputs))
		t.outputs = outputs
		return t
	
	
	def get_signable_form(self, input_index):
		# The transaction-in-signable-form differs for each input.
		s = self.version
		s += self.input_count
		for i, input in enumerate(self.inputs):
			s += input.previous_output_hash
			s += input.previous_output_index
			if i == input_index:
				s += input.script_length_scriptPubKey
				s += input.scriptPubKey
			else:
				# untested - not sure about this. 
				s += "00" # a script_length of 0 that indicates that no script is included. 
			s += input.sequence
		s += self.output_count
		for output in self.outputs:
			# distinguish between p2pkh and p2sh outputs. 
			if hasattr(output, "output_type"):
				if output.output_type == "p2sh":
					s += output.value
					s += output.script_length
					s += output.script
			else:
				# p2pkh output
				s += output.value
				s += output.script_length
				s += output.scriptPubKey
		s += self.block_lock_time
		s += self.hash_type_4_byte
		return s


	def get_description_of_signable_form(self, input_index):
		s = "Nonstandard Transaction (signable form %d): {%d bytes}" % (input_index, bf.hex_len(self.get_signable_form(input_index)))
		s += "\n- version: %s (%d bytes)" % (self.version, bf.hex_len(self.version))
		s += "\n- input_count: %s (%d bytes)" % (self.input_count, bf.hex_len(self.input_count))
		for i, input in enumerate(self.inputs):
			if i == input_index:
				s += "\n- Input %d [to be used to sign this signable form]: {%d bytes}" % (i, bf.hex_len(input.get_signable_form()))
			else:
				s += "\n- Input %d: {%d bytes}" % (i, bf.hex_len(input.get_signable_empty_form()))
			s += "\n-- previous_output_hash: %s (%d bytes)" % (input.previous_output_hash, bf.hex_len(input.previous_output_hash))
			s += "\n-- previous_output_index: %s (%d bytes)" % (input.previous_output_index, bf.hex_len(input.previous_output_index))
			if i == input_index:
				s += "\n-- script_length_scriptPubKey: %s (%d bytes)" % (input.script_length_scriptPubKey, bf.hex_len(input.script_length_scriptPubKey))
				s += "\n-- scriptPubKey: %s (%d bytes)" % (input.scriptPubKey, bf.hex_len(input.scriptPubKey))
			else:
				# untested - not sure about this. 
				s += "\n-- script_length: 00" # # a script_length of 0 that indicates that no script is included.
			s += "\n-- sequence: %s (%d bytes)" % (input.sequence, bf.hex_len(input.sequence))
		s += "\n- output_count: %s (%d bytes)" % (self.output_count, bf.hex_len(self.output_count))
		for i, output in enumerate(self.outputs):
			# distinguish between p2pkh and p2sh outputs. 
			if hasattr(output, "output_type"):
				if output.output_type == "p2sh":
					s += "\n- Output %d (P2SH): {%d bytes}" % (i, bf.hex_len(output.get_form()))
					s += "\n-- value: %s (%d bytes)" % (output.value, bf.hex_len(output.value))
					s += "\n-- script_length: %s (%d bytes)" % (output.script_length, bf.hex_len(output.script_length))
					s += "\n-- script: %s (%d bytes)" % (output.script, bf.hex_len(output.script))
			else:
				# p2pkh output.
				s += "\n- Output %d: {%d bytes}" % (i, bf.hex_len(output.get_form()))
				s += "\n-- value: %s (%d bytes)" % (output.value, bf.hex_len(output.value))
				s += "\n-- script_length: %s (%d bytes)" % (output.script_length, bf.hex_len(output.script_length))
				s += "\n-- scriptPubKey: %s (%d bytes)" % (output.scriptPubKey, bf.hex_len(output.scriptPubKey))
		s += "\n- block_lock_time: %s (%d bytes)" % (self.block_lock_time, bf.hex_len(self.block_lock_time))
		s += "\n- hash_type_4_byte: %s (%d bytes)" % (self.hash_type_4_byte, bf.hex_len(self.hash_type_4_byte))
		return s
		
		
	def sign(self, random_values):
		# only handles signatures made by standard inputs. 
		# future: write a validate_self() instance method, which should be called here. It should validate all the properties of a transaction prior to signing (and the properties of all its inputs and outputs). 
		if len(random_values) != len(self.inputs):
			message = "There must be one random value per input. Each signature of the transaction by an input requires a new random value."
			message += "\nNumber of inputs: %d" % len(inputs)
			message += "\nNumber of random values: %d" % len(random_values)
			bf.stop(message)
		# the domain for a random value for a signature is the same as the domain for bitcoin private keys.
		for value in random_values:
			bf.validate_bitcoin_private_key_hex(value)
		# need to sign once per input, using the private key associated with that input. 
		for i, input in enumerate(self.inputs):
			random_value_hex = random_values[i]
			input_index = i
			signature_hex = self.make_signature_for_one_input(input_index, random_value_hex)
			# Append hash_type SIGHASH_ALL (as a single byte) "01" to DER-encoded signature.
			signature_hex += self.hash_type_1_byte
			bf.validate_hex(signature_hex)
			scriptSig, script_length = 			bf.signature_hex_and_public_key_hex_to_scriptSig(signature_hex, input.public_key_hex)
			input.script_length = script_length
			input.scriptSig = scriptSig
			# storing the scriptSig and its script_length in the input-used-for-signing, for every input, completes the sign() process. 
			
			
	def make_signature_for_one_input(self, input_index, random_value_hex):
		from binascii import hexlify, unhexlify
		input = self.inputs[input_index]
		# get the transaction-in-signable-form for this input
		signable_form_hex = self.get_signable_form(input_index)
		digest_hex = bf.double_sha256(signable_form_hex)
		# get the private_key in raw bytes.
		private_key_bytes = unhexlify(input.private_key_hex)
		# create a SigningKey object from the private-key-in-raw-bytes, specifying that the elliptic curve is secp256k1. 
		signing_key = ecdsa.SigningKey.from_string(private_key_bytes, curve=ecdsa.SECP256k1)
		# confirm that the hash digest byte length is not greater than curve.baselen, where the curve is secp256k1.
		if bf.hex_len(digest_hex) > signing_key.curve.baselen:
			message = "This curve (%s) is too short for signing this digest (%d bytes)." % (signing_key.curve.name, len(digest2))
			bf.stop(message)
		# get double sha256 hash digest as an integer
		digest_int = int(digest_hex, 16)
		# get random_value as an integer. 
		random_value_int = int(random_value_hex, 16)
		# Use the sign_number() method of the SigningKey object to sign the hash digest. Pass random_value_int as an argument for use in the signing process. Retrieve the resulting r and s values, which together constitute the ECDSA signature. 
		# Notes: 
		# - Every signature requires a new and unique random_value. 
		# - r and s are both returned as integers / longs. 
		r, s = signing_key.sign_number(number=digest_int, entropy=None, k=random_value_int)
		# get r and s in hex form
		r_hex = bf.int_to_hex(r)
		s_hex = bf.int_to_hex(s)
		# the domain for both r and s is the same as the domain for bitcoin private keys.
		bf.validate_bitcoin_private_key_hex(r_hex)
		bf.validate_bitcoin_private_key_hex(s_hex)
		# Check that the S-value in the signature is within the low-S domain [1, floor(n/2)]. If it is high-S, alter it to low-S.
		s = bf.ensure_low_s_value(s)
		# Convert the signature (r, s) into DER encoding. The Python ECDSA library includes a function for doing this. 
		signature_bytes = ecdsa.util.sigencode_der(r, s, signing_key.privkey.order)
		signature_hex = hexlify(signature_bytes)
		return signature_hex
	
	
	def get_signed_form(self):
		# Future: Check that the transaction has actually been signed (i.e. that every input has an associated scriptSig).
		s = self.version
		s += self.input_count
		for i, input in enumerate(self.inputs):
			s += input.previous_output_hash
			s += input.previous_output_index
			s += input.script_length
			s += input.scriptSig
			s += input.sequence
		s += self.output_count
		for output in self.outputs:
			# distinguish between p2pkh and p2sh outputs. 
			if hasattr(output, "output_type"):
				if output.output_type == "p2sh":
					s += output.value
					s += output.script_length
					s += output.script
			else:
				# p2pkh output
				s += output.value
				s += output.script_length
				s += output.scriptPubKey
		s += self.block_lock_time
		return s
	
	
	def get_description_of_signed_form(self):
		# Future: Check that the transaction has actually been signed (i.e. that every input has an associated scriptSig).
		s = "Nonstandard Transaction (signed form): {%d bytes}" % bf.hex_len(self.get_signed_form())
		s += "\n- version: %s (%d bytes)" % (self.version, bf.hex_len(self.version))
		s += "\n- input_count: %s (%d bytes)" % (self.input_count, bf.hex_len(self.input_count))
		for i, input in enumerate(self.inputs):
			s += "\n- Input %d: {%d bytes}" % (i, bf.hex_len(input.get_signed_form()))
			s += "\n-- previous_output_hash: %s (%d bytes)" % (input.previous_output_hash, bf.hex_len(input.previous_output_hash))
			s += "\n-- previous_output_index: %s (%d bytes)" % (input.previous_output_index, bf.hex_len(input.previous_output_index))
			s += "\n-- script_length: %s (%d bytes)" % (input.script_length, bf.hex_len(input.script_length))
			s += "\n-- scriptSig: %s (%d bytes)" % (input.scriptSig, bf.hex_len(input.scriptSig))
			s += "\n-- sequence: %s (%d bytes)" % (input.sequence, bf.hex_len(input.sequence))
		s += "\n- output_count: %s (%d bytes)" % (self.output_count, bf.hex_len(self.output_count))
		for i, output in enumerate(self.outputs):
			# distinguish between p2pkh and p2sh outputs. 
			if hasattr(output, "output_type"):
				if output.output_type == "p2sh":
					s += "\n- Output %d (P2SH): {%d bytes}" % (i, bf.hex_len(output.get_form()))
					s += "\n-- value: %s (%d bytes)" % (output.value, bf.hex_len(output.value))
					s += "\n-- script_length: %s (%d bytes)" % (output.script_length, bf.hex_len(output.script_length))
					s += "\n-- script: %s (%d bytes)" % (output.script, bf.hex_len(output.script))
			else:
				# p2pkh output.
				s += "\n- Output %d: {%d bytes}" % (i, bf.hex_len(output.get_form()))
				s += "\n-- value: %s (%d bytes)" % (output.value, bf.hex_len(output.value))
				s += "\n-- script_length: %s (%d bytes)" % (output.script_length, bf.hex_len(output.script_length))
				s += "\n-- scriptPubKey: %s (%d bytes)" % (output.scriptPubKey, bf.hex_len(output.scriptPubKey))
		s += "\n- block_lock_time: %s (%d bytes)" % (self.block_lock_time, bf.hex_len(self.block_lock_time))
		return s
		
	
	



class P2SHOutput:


	def __init__(self):
		# properties that go into the raw transaction:
		self.value = None # 8 bytes (little-endian)
		self.script_length = None # var_int
		self.script = None # hex bytes
		# ^ this script is the outputScript. This takes the place of the scriptPubKey in a P2PKH Output. 
		# other properties:
		self.address = None # string
		self.bitcoin_amount = None # string
		self.satoshi_amount = None # int
		self.redeem_script_hash_hex = None # string (20 hex bytes)
		self.output_type = "p2sh"
	
	
	def get_form(self):
		s = self.value
		s += self.script_length
		s += self.script
		return s
		
		
	def get_description(self):
		s = "P2SH Output: {%d bytes}" % (bf.hex_len(self.get_form()))
		s += "\n- [data] value: %s (%d bytes)" % (self.value, bf.hex_len(self.value))
		s += "\n- [data] script_length: %s (%d bytes)" % (self.script_length, bf.hex_len(self.script_length))
		s += "\n- [data] script: %s (%d bytes)" % (self.script, bf.hex_len(self.script))
		s += "\n- [more info] address: %s" % self.address
		s += "\n- [more info] bitcoin_amount: %s" % self.bitcoin_amount
		s += "\n- [more info] satoshi_amount: %s" % self.satoshi_amount
		s += "\n- [more info] redeem_script_hash_hex: %s" % self.redeem_script_hash_hex
		return s
		
		
	@classmethod
	def create(cls, value, script, script_length):
		o = P2SHOutput()
		o.value = value
		o.script = script
		o.script_length = script_length
		return o
	
	
	@classmethod
	def create_from_address(cls, address, bitcoin_amount):
		# calculate more fundamental properties from the inputs.
		satoshi_amount = bf.bitcoin_to_satoshi(bitcoin_amount)
		value = bf.satoshi_to_tx_output_hex_le(satoshi_amount)
		hash_hex = nbf.p2sh_address_to_redeem_script_hash_hex(address)
		script, script_length = nbf.redeem_script_hash_hex_to_script(hash_hex)
		# call the more basic create() method. 
		o = P2SHOutput.create(value, script, script_length)
		o.address = address
		o.bitcoin_amount = bitcoin_amount
		o.satoshi_amount = satoshi_amount
		o.redeem_script_hash_hex = hash_hex
		return o
		
		
	def set_satoshi_amount(self, satoshi_amount):
		bf.validate_positive_int(satoshi_amount)
		self.satoshi_amount = satoshi_amount
		self.bitcoin_amount = bf.satoshi_to_bitcoin(satoshi_amount)
		self.value = bf.satoshi_to_tx_output_hex_le(satoshi_amount)
	
	