# DETAILS
# - Author: StJohn Piano
# - Date: 2019-02-23
# - Description: This file contains classes for a standard Bitcoin transaction, a standard Bitcoin transaction input, and a standard Bitcoin transaction output. 


# ENVIRONMENTS
# - Tested on:
# -- Python 2.7.13 running on Mac OS X 10.6.8 (Snow Leopard). 


# NOTES
# 
# - My working definition of a standard transaction: 
# -- It has at least one input and at least one output. 
# -- All input and output addresses are Pay-To-Public-Key-Hash (P2PKH). 
# -- All input scriptSigs contain uncompressed public keys.


# DEPENDENCIES
import bitcoin_functions_2 as bf
import ecdsa






class Transaction:

	# This is my current working definition of the raw format of a standard Bitcoin transaction:
	#
	# Standard transaction:
	# START FORMAT
	# - version: 4 bytes (little-endian)
	# - input_count: (var_int)
	# - [for each input:]
	# -- previous_output_hash: 32 bytes (big-endian). When little-endian, it is the "txid". 
	# -- previous_output_index: 4 bytes (little-endian)
	# -- script_length: (var_int)
	# -- scriptSig [see below]
	# -- sequence: 4 bytes (little-endian)
	# - output_count: (var_int)
	# - [for each output:]
	# -- value: 8 bytes (little-endian)
	# -- script_length: (var_int)
	# -- scriptPubKey [see below]
	# - block lock time: 4 bytes
	# END FORMAT
	#
	# scriptSig:
	# START FORMAT
	# - PUSHDATA: 47 (approximately)
	# - [derived property] PUSHDATA decimal value: 71 (approximately)
	# - signature_data: 71 bytes (approximately)
	# - PUSHDATA: 41
	# - [derived property] PUSHDATA decimal value: 65
	# - public_key_data: 65 bytes ("04", 32-byte big-endian X value, 32-byte big-endian Y value)
	# END FORMAT
	# 
	# scriptPubKey:
	# START FORMAT
	# - OP_DUP: 76
	# - OP_HASH160: a9
	# - PUSHDATA: 14
	# - [derived property] PUSHDATA decimal value: 20
	# - public_key_hash: 20 bytes (big-endian)
	# - OP_EQUALVERIFY: 88
	# - OP_CHECKSIG: ac
	# END FORMAT
	#
	# Notes:
	# - var_int = (variable-length byte sequence)
	# - I think that the version is always hardcoded as '1' for the main Bitcoin network. In 4 little-endian hex bytes, this is 0x01 00 00 00. 
	# - [Hearsay] Sequence can always be set to 0xff ff ff ff. It is a nonfunctional legacy feature that was designed to allow multiple signers to agree to update a transaction. 
	# - [Hearsay] Block lock time can always be set to 0x00 00 00 00. It was designed to ask miners not to mine a transaction until a specified block height is reached. A block lock time of 0x00 00 00 00 specifies that the transaction can be mined at any block height after 0. 
	# - The previous_output_hash is:
	# -- the hash of the previous transaction that:
	# --- contains the unspent output that:
	# ---- will be used as an input in this new transaction
	# - [Hearsay] The hash type (SIGHASH_ALL) is 1. As 1 byte, this is 0x01. As 4 bytes (little-endian), this is 0x01 00 00 00. The 4-byte version is appended to the transaction-in-signable-form before signing. The 1-byte version is appended to the signature_data after signing.
	#
	#
	# This is my current working definition of the raw format of a standard Bitcoin transaction-in-signable-form:
	# 
	# START FORMAT
	# - version: 4 bytes (little-endian)
	# - input_count: (var_int)
	# - [for each input:]
	# -- previous_output_hash: 32 bytes (big-endian)
	# -- previous_output_index: 4 bytes (little-endian)
	# -- script_length: (var_int)
	# -- scriptPubKey of relevant unspent output from previous transaction [same format as scriptPubKey - see below]
	# -- sequence: 4 bytes (little-endian)
	# - output_count: (var_int)
	# - [for each output:]
	# -- value: 8 bytes (little-endian)
	# -- script_length: (var_int)
	# -- scriptPubKey [see below]
	# - block lock time: 4 bytes
	# - hash type: 4 bytes (little-endian)
	# END FORMAT
	# 
	# scriptPubKey (standard form):
	# START FORMAT
	# - OP_DUP: 76
	# - OP_HASH160: a9
	# - PUSHDATA: 14
	# - [derived property] PUSHDATA decimal value: 20
	# - public_key_hash: 20 bytes (big-endian)
	# - OP_EQUALVERIFY: 88
	# - OP_CHECKSIG: ac
	# END FORMAT
	# 
	# Note: In the {scriptPubKey of relevant unspent output from previous transaction}, the unspent output is associated with a particular address, the private key of which is to be used to sign this particular transaction-in-signable-form in order to create a single scriptSig. 
	

	def __init__(self):
		# properties that go into the raw transaction:
		self.version = "01000000"
		self.input_count = None # var_int
		self.inputs = []
		self.output_count = None # var_int
		self.outputs = []
		self.block_lock_time = "00000000"
		# properties that go into the raw transaction-in-signable-form:
		self.hash_type_4_byte = "01000000" # 4 bytes (little-endian)
		# properties that go into the signed transaction:
		self.hash_type_1_byte = "01"
		
		
	def get_unsigned_form(self):
		# This is the basic transaction, with no signatures.
		s = self.version
		s += self.input_count
		for i, input in enumerate(self.inputs):
			s += input.previous_output_hash
			s += input.previous_output_index
			# input.script_length not included
			# input.scriptSig not included
			s += input.sequence
		s += self.output_count
		for output in self.outputs:
			s += output.value
			s += output.script_length
			s += output.scriptPubKey
		s += self.block_lock_time
		return s
	
	
	def get_description_of_unsigned_form(self):
		s = "Transaction (unsigned form): {%d bytes}" % bf.hex_len(self.get_unsigned_form())
		s += "\n- version: %s (%d bytes)" % (self.version, bf.hex_len(self.version))
		s += "\n- input_count: %s (%d bytes)" % (self.input_count, bf.hex_len(self.input_count))
		for i, input in enumerate(self.inputs):
			s += "\n- Input %d: {%d bytes}" % (i, bf.hex_len(input.get_unsigned_form()))
			s += "\n-- previous_output_hash: %s (%d bytes)" % (input.previous_output_hash, bf.hex_len(input.previous_output_hash))
			s += "\n-- previous_output_index: %s (%d bytes)" % (input.previous_output_index, bf.hex_len(input.previous_output_index))
			s += "\n-- script_length: %s (%d bytes)" % ("[none]", 0)
			s += "\n-- scriptSig: %s (%d bytes)" % ("[none]", 0)
			s += "\n-- sequence: %s (%d bytes)" % (input.sequence, bf.hex_len(input.sequence))
		s += "\n- output_count: %s (%d bytes)" % (self.output_count, bf.hex_len(self.output_count))
		for i, output in enumerate(self.outputs):
			s += "\n- Output %d: {%d bytes}" % (i, bf.hex_len(output.get_form()))
			s += "\n-- value: %s (%d bytes)" % (output.value, bf.hex_len(output.value))
			s += "\n-- script_length: %s (%d bytes)" % (output.script_length, bf.hex_len(output.script_length))
			s += "\n-- scriptPubKey: %s (%d bytes)" % (output.scriptPubKey, bf.hex_len(output.scriptPubKey))
		s += "\n- block_lock_time: %s (%d bytes)" % (self.block_lock_time, bf.hex_len(self.block_lock_time))
		return s
	
	
	@classmethod
	def create(cls, inputs, outputs):
		t = Transaction()
		t.input_count = bf.int_to_var_int(len(inputs))
		t.inputs = inputs
		t.output_count = bf.int_to_var_int(len(outputs))
		t.outputs = outputs
		return t
	
	
	def get_signable_form(self, input_index):
		# The transaction-in-signable-form differs for each input.
		s = self.version
		s += self.input_count
		for i, input in enumerate(self.inputs):
			s += input.previous_output_hash
			s += input.previous_output_index
			if i == input_index:
				s += input.script_length_scriptPubKey
				s += input.scriptPubKey
			else:
				# untested - not sure about this. 
				s += "00" # a script_length of 0 that indicates that no script is included. 
			s += input.sequence
		s += self.output_count
		for output in self.outputs:
			s += output.value
			s += output.script_length
			s += output.scriptPubKey
		s += self.block_lock_time
		s += self.hash_type_4_byte
		return s
	
	
	def get_description_of_signable_form(self, input_index):
		s = "Transaction (signable form %d): {%d bytes}" % (input_index, bf.hex_len(self.get_signable_form(input_index)))
		s += "\n- version: %s (%d bytes)" % (self.version, bf.hex_len(self.version))
		s += "\n- input_count: %s (%d bytes)" % (self.input_count, bf.hex_len(self.input_count))
		for i, input in enumerate(self.inputs):
			if i == input_index:
				s += "\n- Input %d [to be used to sign this signable form]: {%d bytes}" % (i, bf.hex_len(input.get_signable_form()))
			else:
				s += "\n- Input %d: {%d bytes}" % (i, bf.hex_len(input.get_signable_empty_form()))
			s += "\n-- previous_output_hash: %s (%d bytes)" % (input.previous_output_hash, bf.hex_len(input.previous_output_hash))
			s += "\n-- previous_output_index: %s (%d bytes)" % (input.previous_output_index, bf.hex_len(input.previous_output_index))
			if i == input_index:
				s += "\n-- script_length_scriptPubKey: %s (%d bytes)" % (input.script_length_scriptPubKey, bf.hex_len(input.script_length_scriptPubKey))
				s += "\n-- scriptPubKey: %s (%d bytes)" % (input.scriptPubKey, bf.hex_len(input.scriptPubKey))
			else:
				# untested - not sure about this. 
				s += "\n-- script_length: 00 (1 bytes)" # a script_length of 0 that indicates that no script is included.
			s += "\n-- sequence: %s (%d bytes)" % (input.sequence, bf.hex_len(input.sequence))
		s += "\n- output_count: %s (%d bytes)" % (self.output_count, bf.hex_len(self.output_count))
		for i, output in enumerate(self.outputs):
			s += "\n- Output %d: {%d bytes}" % (i, bf.hex_len(output.get_form()))
			s += "\n-- value: %s (%d bytes)" % (output.value, bf.hex_len(output.value))
			s += "\n-- script_length: %s (%d bytes)" % (output.script_length, bf.hex_len(output.script_length))
			s += "\n-- scriptPubKey: %s (%d bytes)" % (output.scriptPubKey, bf.hex_len(output.scriptPubKey))
		s += "\n- block_lock_time: %s (%d bytes)" % (self.block_lock_time, bf.hex_len(self.block_lock_time))
		s += "\n- hash_type_4_byte: %s (%d bytes)" % (self.hash_type_4_byte, bf.hex_len(self.hash_type_4_byte))
		return s
	
	
	def sign(self, random_values):
		# future: write a validate_self() instance method, which should be called here. It should validate all the properties of a transaction prior to signing (and the properties of all its inputs and outputs). 
		if len(random_values) != len(self.inputs):
			message = "There must be one random value per input. Each signature of the transaction by an input requires a new random value."
			message += "\nNumber of inputs: %d" % len(inputs)
			message += "\nNumber of random values: %d" % len(random_values)
			bf.stop(message)
		# the domain for a random value for a signature is the same as the domain for bitcoin private keys.
		for value in random_values:
			bf.validate_bitcoin_private_key_hex(value)
		# need to sign once per input, using the private key associated with that input. 
		for i, input in enumerate(self.inputs):
			random_value_hex = random_values[i]
			input_index = i
			signature_hex = self.make_signature_for_one_input(input_index, random_value_hex)
			# Append hash_type SIGHASH_ALL (as a single byte) "01" to DER-encoded signature.
			signature_hex += self.hash_type_1_byte
			bf.validate_hex(signature_hex)
			scriptSig, script_length = bf.signature_hex_and_public_key_hex_to_scriptSig(signature_hex, input.public_key_hex)
			input.script_length = script_length
			input.scriptSig = scriptSig
			# storing the scriptSig and its script_length in the input-used-for-signing, for every input, completes the sign() process. 
	
	
	def make_signature_for_one_input(self, input_index, random_value_hex):
		from binascii import hexlify, unhexlify
		input = self.inputs[input_index]
		# get the transaction-in-signable-form for this input
		signable_form_hex = self.get_signable_form(input_index)
		# calculate the hash digest SHA256(SHA256(transaction-in-signable-form))
		digest_hex = bf.double_sha256(signable_form_hex)
		bf.validate_hex_length(digest_hex, 32)
		# get the private_key in raw bytes.
		private_key_bytes = unhexlify(input.private_key_hex)
		# create a SigningKey object from the private-key-in-raw-bytes, specifying that the elliptic curve is secp256k1. 
		signing_key = ecdsa.SigningKey.from_string(private_key_bytes, curve=ecdsa.SECP256k1)
		# confirm that the hash digest byte length is not greater than curve.baselen, where the curve is secp256k1.
		if bf.hex_len(digest_hex) > signing_key.curve.baselen:
			message = "This curve (%s) is too short for signing this digest (%d bytes)." % (signing_key.curve.name, bf.hex_len(digest_hex))
			bf.stop(message)
		# get hash digest as an integer
		digest_int = int(digest_hex, 16)
		# get random_value as an integer. 
		random_value_int = int(random_value_hex, 16)
		# Use the sign_number() method of the SigningKey object to sign the hash digest. Pass random_value_int as an argument for use in the signing process. Retrieve the resulting r and s values, which together constitute the ECDSA signature. 
		# Notes: 
		# - Every signature requires a new and unique random_value. 
		# - r and s are both returned as integers / longs. 
		r, s = signing_key.sign_number(number=digest_int, entropy=None, k=random_value_int)
		# get r and s in hex form
		r_hex = bf.int_to_hex(r)
		s_hex = bf.int_to_hex(s)
		# the domain for both r and s is the same as the domain for bitcoin private keys.
		bf.validate_bitcoin_private_key_hex(r_hex)
		bf.validate_bitcoin_private_key_hex(s_hex)
		# Check that the S-value in the signature is within the low-S domain [1, floor(n/2)]. If it is high-S, alter it to low-S. 
		s = bf.ensure_low_s_value(s)
		# Convert the signature (r, s) into DER encoding. The Python ECDSA library includes a function for doing this. 
		signature_bytes = ecdsa.util.sigencode_der(r, s, signing_key.privkey.order)
		signature_hex = hexlify(signature_bytes)
		return signature_hex
	
	
	def get_signed_form(self):
		# Future: Check that the transaction has actually been signed (i.e. that every input has an associated scriptSig).
		s = self.version
		s += self.input_count
		for i, input in enumerate(self.inputs):
			s += input.previous_output_hash
			s += input.previous_output_index
			s += input.script_length
			s += input.scriptSig
			s += input.sequence
		s += self.output_count
		for output in self.outputs:
			s += output.value
			s += output.script_length
			s += output.scriptPubKey
		s += self.block_lock_time
		return s
	
	
	def get_description_of_signed_form(self):
		# Future: Check that the transaction has actually been signed (i.e. that every input has an associated scriptSig).
		s = "Transaction (signed form): {%d bytes}" % bf.hex_len(self.get_signed_form())
		s += "\n- version: %s (%d bytes)" % (self.version, bf.hex_len(self.version))
		s += "\n- input_count: %s (%d bytes)" % (self.input_count, bf.hex_len(self.input_count))
		for i, input in enumerate(self.inputs):
			s += "\n- Input %d: {%d bytes}" % (i, bf.hex_len(input.get_signed_form()))
			s += "\n-- previous_output_hash: %s (%d bytes)" % (input.previous_output_hash, bf.hex_len(input.previous_output_hash))
			s += "\n-- previous_output_index: %s (%d bytes)" % (input.previous_output_index, bf.hex_len(input.previous_output_index))
			s += "\n-- script_length: %s (%d bytes)" % (input.script_length, bf.hex_len(input.script_length))
			s += "\n-- scriptSig: %s (%d bytes)" % (input.scriptSig, bf.hex_len(input.scriptSig))
			s += "\n-- sequence: %s (%d bytes)" % (input.sequence, bf.hex_len(input.sequence))
		s += "\n- output_count: %s (%d bytes)" % (self.output_count, bf.hex_len(self.output_count))
		for i, output in enumerate(self.outputs):
			s += "\n- Output %d: {%d bytes}" % (i, bf.hex_len(output.get_form()))
			s += "\n-- value: %s (%d bytes)" % (output.value, bf.hex_len(output.value))
			s += "\n-- script_length: %s (%d bytes)" % (output.script_length, bf.hex_len(output.script_length))
			s += "\n-- scriptPubKey: %s (%d bytes)" % (output.scriptPubKey, bf.hex_len(output.scriptPubKey))
		s += "\n- block_lock_time: %s (%d bytes)" % (self.block_lock_time, bf.hex_len(self.block_lock_time))
		return s
	





class Output:
	
	
	def __init__(self):
		# properties that go into the raw transaction:
		self.value = None # 8 bytes (little-endian)
		self.script_length = None # var_int
		self.scriptPubKey = None # hex bytes
		# other properties:
		self.address = None # string
		self.bitcoin_amount = None # string
		self.satoshi_amount = None # int
		self.public_key_hash_hex = None # string (20 hex bytes)
	
	
	def get_form(self):
		s = self.value
		s += self.script_length
		s += self.scriptPubKey
		return s
	
	
	def get_description(self):
		s = "Output: {%d bytes}" % (bf.hex_len(self.get_form()))
		s += "\n- [data] value: %s (%d bytes)" % (self.value, bf.hex_len(self.value))
		s += "\n- [data] script_length: %s (%d bytes)" % (self.script_length, bf.hex_len(self.script_length))
		s += "\n- [data] scriptPubKey: %s (%d bytes)" % (self.scriptPubKey, bf.hex_len(self.scriptPubKey))
		s += "\n- [more info] address: %s" % self.address
		s += "\n- [more info] bitcoin_amount: %s" % self.bitcoin_amount
		s += "\n- [more info] satoshi_amount: %s" % self.satoshi_amount
		s += "\n- [more info] public_key_hash_hex: %s" % self.public_key_hash_hex
		return s
	
	
	@classmethod
	def create(cls, value, scriptPubKey, script_length):
		o = Output()
		o.value = value
		o.scriptPubKey = scriptPubKey
		o.script_length = script_length
		return o
	
	
	@classmethod
	def create_from_address(cls, address, bitcoin_amount):
		# calculate more fundamental properties from the inputs.
		satoshi_amount = bf.bitcoin_to_satoshi(bitcoin_amount)
		value = bf.satoshi_to_tx_output_hex_le(satoshi_amount)
		hash_hex = bf.bitcoin_address_to_public_key_hash_hex(address)
		scriptPubKey, script_length = bf.public_key_hash_hex_to_scriptPubKey(hash_hex)
		# call the more basic create() method. 
		o = Output.create(value, scriptPubKey, script_length)
		o.address = address
		o.bitcoin_amount = bitcoin_amount
		o.satoshi_amount = satoshi_amount
		o.public_key_hash_hex = hash_hex
		return o
	
	
	def set_satoshi_amount(self, satoshi_amount):
		bf.validate_positive_int(satoshi_amount)
		self.satoshi_amount = satoshi_amount
		self.bitcoin_amount = bf.satoshi_to_bitcoin(satoshi_amount)
		self.value = bf.satoshi_to_tx_output_hex_le(satoshi_amount)
	
	




class Input:


	def __init__(self):
		# properties that go into the raw transaction:
		self.previous_output_hash = None # 32 bytes (big-endian). When little-endian, it is the "txid". 
		self.previous_output_index = None # 4 bytes (little-endian)
		self.sequence = None # 4 bytes
		# properties that are needed for signing the raw transaction:
		self.private_key_hex = None # 32 bytes
		self.public_key_hex = None # Included in the scriptSig. Used when validating transaction signatures. 
		self.script_length_scriptPubKey = None # var_int. Length of the scriptPubKey.
		self.scriptPubKey = None # The "scriptPubKey of relevant unspent output from previous transaction" (roughly speaking: the address from which this input originates) is included in the signable form of a transaction when signing that transaction with the particular input. 
		# properties that go into the signed transaction:
		self.script_length = None
		self.scriptSig = None
		# other properties:
		self.address = None # source address (derived from self.private_key_hex)
		self.previous_output_index_int = None # string. integer value of index. 
		self.txid = None # little-endian form of previous_output_hash. 
		self.satoshi_amount = None # int
		self.bitcoin_amount = None # string
	
	
	def get_unsigned_form(self):
		s = self.previous_output_hash
		s += self.previous_output_index
		s += self.sequence
		return s
	
	
	def get_signable_form(self):
		s = self.previous_output_hash
		s += self.previous_output_index
		s += self.script_length_scriptPubKey
		s += self.scriptPubKey
		s += self.sequence
		return s
	
	
	def get_signable_empty_form(self):
		# this is the form used when another input is being used to sign the transaction.
		s = self.previous_output_hash
		s += self.previous_output_index
		s += bf.int_to_var_int(0) # 0x00 script_length for scriptSig
		# no scriptSig
		s += self.sequence
		return s 		
	
	
	def get_signed_form(self):
		s = self.previous_output_hash
		s += self.previous_output_index
		if self.script_length == None or self.scriptSig == None:
			message = "This input has not been signed yet."
			bf.stop(message)
		s += self.script_length
		s += self.scriptSig
		s += self.sequence
		return s
	
	
	def get_description(self):
		s = "Input: "
		s += "{%d bytes unsigned" % bf.hex_len(self.get_unsigned_form())
		s += ", %d bytes signable" % bf.hex_len(self.get_signable_form())
		if self.scriptSig != None:
			s += ", %d bytes signed" % bf.hex_len(self.get_signed_form())
		s += "}"
		s += "\n- [data] previous_output_hash: %s (%d bytes)" % (self.previous_output_hash, bf.hex_len(self.previous_output_hash))
		s += "\n- [data] previous_output_index: %s (%d bytes)" % (self.previous_output_index, bf.hex_len(self.previous_output_index))
		s += "\n- [signed form] script_length: "
		if self.script_length == None: 
			s += "[none] (0 bytes)"
		else:
			s += "%s (%d bytes)" % (self.script_length, bf.hex_len(self.script_length))
		s += "\n- [signed form] scriptSig: "
		if self.scriptSig == None: 
			s += "[none] (0 bytes)"
		else:
			s += "%s (%d bytes)" % (self.scriptSig, bf.hex_len(self.scriptSig))
		s += "\n- [data] sequence: %s (%d bytes)" % (self.sequence, bf.hex_len(self.sequence))
		s += "\n- [used for signing] private_key_hex: %s (%d bytes)" % (self.private_key_hex, bf.hex_len(self.private_key_hex))
		s += "\n- [source, goes into scriptSig] public_key_hex: %s (%d bytes)" % (self.public_key_hex, bf.hex_len(self.public_key_hex))
		s += "\n- [signable form] script_length_scriptPubKey: %s (%d bytes)" % (self.script_length_scriptPubKey, bf.hex_len(self.script_length_scriptPubKey))
		s += "\n- [signable form] scriptPubKey: %s (%d bytes)" % (self.scriptPubKey, bf.hex_len(self.scriptPubKey))
		s += "\n- [more info] address: %s" % self.address
		s += "\n- [more info] previous_output_index_int: %s" % self.previous_output_index_int
		s += "\n- [more info] txid: %s" % self.txid
		return s
	
	
	@classmethod
	def create(cls, previous_output_hash, previous_output_index, private_key_hex, public_key_hex, scriptPubKey, script_length_scriptPubKey):
		bf.validate_hex_length(previous_output_hash, 32)
		bf.validate_hex_length(previous_output_index, 4)
		bf.validate_bitcoin_private_key_hex(private_key_hex)
		bf.validate_hex(public_key_hex)
		bf.validate_hex(scriptPubKey)
		bf.validate_hex(script_length_scriptPubKey)
		i = Input()
		i.previous_output_hash = previous_output_hash
		i.previous_output_index = previous_output_index
		i.private_key_hex = private_key_hex
		i.public_key_hex = public_key_hex
		i.scriptPubKey = scriptPubKey
		i.script_length_scriptPubKey = script_length_scriptPubKey
		i.sequence = "ffffffff"
		return i
	
	
	@classmethod
	def create_from_txid(cls, txid, previous_output_index, private_key_hex, bitcoin_amount):
		previous_output_hash = bf.reverse_hex_order(txid)
		# pad private_key_hex to 32 bytes
		private_key_hex = bf.left_pad_hex_to_length_x(private_key_hex, 32)
		bf.validate_hex_length(private_key_hex, 32)
		# previous_output_index is an integer. needs conversion to 4 hex bytes (little endian). 
		bf.validate_int(previous_output_index)
		previous_output_index_int = previous_output_index
		previous_output_index = bf.tx_index_int_to_tx_index_hex_le(previous_output_index)
		# get public_key_hex from private_key_hex
		public_key_hex = bf.private_key_hex_to_public_key_hex(private_key_hex)
		bf.validate_hex(public_key_hex)
		# get scriptPubKey from public_key_hex.
		hash_hex = bf.public_key_hex_to_hash_hex(public_key_hex)
		bf.validate_hex_length(hash_hex, 20)
		scriptPubKey, script_length_scriptPubKey = bf.public_key_hash_hex_to_scriptPubKey(hash_hex)
		# get bitcoin address from private_key_hex
		address = bf.private_key_hex_to_bitcoin_address(private_key_hex)
		# calculate satoshi amount from bitcoin amount.
		satoshi_amount = bf.bitcoin_to_satoshi(bitcoin_amount)
		# call the more basic create() method. 
		i = Input.create(previous_output_hash, previous_output_index, private_key_hex, public_key_hex, scriptPubKey, script_length_scriptPubKey)
		i.address = address
		i.previous_output_index_int = previous_output_index_int
		i.txid = txid
		i.satoshi_amount = satoshi_amount
		i.bitcoin_amount = bitcoin_amount
		return i






